@@ -6,19 +6,14 @@
 #include <X11/Xlibint.h>
 
 #include "xklavier_private.h"
-#include "xklavier_private_xkb.h"
 
 int XklFilterEvents( XEvent * xev )
 {
   XAnyEvent *pe = ( XAnyEvent * ) xev;
   XklDebug( 400, "**> Filtering event %d of type %d from window %d\n",
             pe->serial, pe->type, pe->window );
-#ifdef XKB_HEADERS_PRESENT
-  if( xev->type == _xklXkbEventType )
-  {
-    _XklXkbEvHandler( ( XkbEvent * ) xev );
-  } else
-#endif
+  _XklEnsureVTableInited();
+  if ( !xklVTable->xklEventHandler( xev ) )
     switch ( xev->type )
     {                           /* core events */
       case FocusIn:
@@ -38,27 +33,26 @@ int XklFilterEvents( XEvent * xev )
                   xev->xdestroywindow.window );
         break;
       case UnmapNotify:
-        XklDebug( 200, "UnmapNotify\n" );
+        XklDebug( 200, "%s\n",  _XklGetEventName( xev->type ) );
         break;
       case MapNotify:
-        XklDebug( 200, "MapNotify\n" );
+        XklDebug( 200, "%s\n",  _XklGetEventName( xev->type ) );
         break;
       case MappingNotify:
-        XklDebug( 200, "MappingNotify\n" );
+        XklDebug( 200, "%s\n",  _XklGetEventName( xev->type ) );
         _XklFreeAllInfo(  );
         _XklLoadAllInfo(  );
         break;
       case GravityNotify:
-        XklDebug( 200, "GravityNotify\n" );
+        XklDebug( 200, "%s\n",  _XklGetEventName( xev->type ) );
         break;
       case ReparentNotify:
-        XklDebug( 200, "ReparentNotify\n" );
+        XklDebug( 200, "%s\n",  _XklGetEventName( xev->type ) );
         break;                  /* Ignore these events */
       default:
       {
-        const char *name = _XklGetEventName( xev->type );
-        XklDebug( 200, "Unknown event %d [%s]\n", xev->type,
-                  ( name == NULL ? "??" : name ) );
+        XklDebug( 200, "Unknown event %d [%s]\n", 
+                       xev->type, _XklGetEventName( xev->type ) );
         return 1;
       }
     }
@@ -67,154 +61,6 @@ int XklFilterEvents( XEvent * xev )
   return 1;
 }
 
-/**
- * Some common functionality for Xkb handler
- */
-void _XklStdXkbHandler( int grp, XklStateChange changeType, unsigned inds,
-                        Bool setInds )
-{
-  Window focused, focusedApp;
-  XklState oldState;
-  int revert;
-  Bool haveState;
-  Bool setGroup = changeType == GROUP_CHANGED;
-
-  XGetInputFocus( _xklDpy, &focused, &revert );
-
-  if( ( focused == None ) || ( focused == PointerRoot ) )
-  {
-    XklDebug( 160, "Something with focus: " WINID_FORMAT "\n", focused );
-    return;
-  }
-
-  if( !_XklGetAppWindow( focused, &focusedApp ) )
-    focusedApp = _xklCurClient; //what else can I do
-
-  XklDebug( 150, "Focused window: " WINID_FORMAT ", '%s'\n", focusedApp,
-            _XklGetDebugWindowTitle( focusedApp ) );
-  XklDebug( 150, "CurClient: " WINID_FORMAT ", '%s'\n", _xklCurClient,
-            _XklGetDebugWindowTitle( _xklCurClient ) );
-
-  if( focusedApp != _xklCurClient )
-  {
-    if ( !_XklGetAppState( focusedApp, &oldState ) )
-    {
-      _XklUpdateCurState( grp, inds, 
-                          "Updating the state from new focused window" );
-      _XklAddAppWindow( focusedApp, ( Window ) NULL, False, &_xklCurState );
-    }
-    else
-    {
-      grp = oldState.group;
-      inds = oldState.indicators;
-    }
-    _xklCurClient = focusedApp;
-    XklDebug( 160, "CurClient:changed to " WINID_FORMAT ", '%s'\n",
-              _xklCurClient, _XklGetDebugWindowTitle( _xklCurClient ) );
-  }
-  // if the window already has this this state - we are just restoring it!
-  // (see the second parameter of stateCallback
-  haveState = _XklGetAppState( _xklCurClient, &oldState );
-
-  if( setGroup || haveState )
-  {
-    _XklUpdateCurState( setGroup ? grp : oldState.group,
-                        setInds ? inds : oldState.indicators,
-                        "Restoring the state from the window" );
-  }
-
-  if( haveState )
-    _XklTryCallStateCallback( changeType, &oldState );
-
-  _XklSaveAppState( _xklCurClient, &_xklCurState );
-}
-
-#ifdef XKB_HEADERS_PRESENT
-/**
- * XKB event handler
- */
-void _XklXkbEvHandler( XkbEvent * kev )
-{
-  int i;
-  unsigned bit;
-  unsigned inds;
-
-  XklDebug( 150, "Xkb event detected\n" );
-
-  switch ( kev->any.xkb_type )
-  {
-    case XkbStateNotify:
-#define GROUP_CHANGE_MASK \
-    ( XkbGroupStateMask | XkbGroupBaseMask | XkbGroupLatchMask | XkbGroupLockMask )
-
-      XklDebug( 150,
-                "XkbStateNotify detected, changes: %X/(mask %X), new group %d\n",
-                kev->state.changed, GROUP_CHANGE_MASK,
-                kev->state.locked_group );
-
-      if( kev->state.changed & GROUP_CHANGE_MASK )
-        _XklStdXkbHandler( kev->state.locked_group, GROUP_CHANGED, 0, False );
-      else
-      {
-        XklDebug( 200,
-                  "This type of state notification is not regarding groups\n" );
-        if ( kev->state.locked_group != _xklCurState.group )
-          XklDebug( 0, 
-                    "ATTENTION! Currently cached group %d is not equal to the current group from the event: %d\n!",
-                    _xklCurState.group,
-                    kev->state.locked_group );
-      }
-
-      break;
-
-    case XkbIndicatorStateNotify:
-
-      XklDebug( 150, "XkbIndicatorStateNotify\n" );
-
-      inds = _xklCurState.indicators;
-
-      ForPhysIndicators( i, bit ) if( kev->indicators.changed & bit )
-      {
-        if( kev->indicators.state & bit )
-          inds |= bit;
-        else
-          inds &= ~bit;
-      }
-
-      _XklStdXkbHandler( 0, INDICATORS_CHANGED, inds, True );
-      break;
-
-    case XkbIndicatorMapNotify:
-      XklDebug( 150, "XkbIndicatorMapNotify\n" );
-      _XklFreeAllInfo(  );
-      _XklLoadAllInfo(  );
-      break;
-
-    case XkbControlsNotify:
-      XklDebug( 150, "XkbControlsNotify\n" );
-      _XklFreeAllInfo(  );
-      _XklLoadAllInfo(  );
-      break;
-
-    case XkbNamesNotify:
-      XklDebug( 150, "XkbNamesNotify\n" );
-      _XklFreeAllInfo(  );
-      _XklLoadAllInfo(  );
-      break;
-
-    case XkbNewKeyboardNotify:
-      XklDebug( 150, "XkbNewKeyboardNotify\n" );
-      _XklFreeAllInfo(  );
-      _XklLoadAllInfo(  );
-      break;
-
-    default:
-      XklDebug( 150, "Unknown xkb event %d\n", kev->any.xkb_type );
-      break;
-  }
-}
-#endif
-
 /**
  * FocusIn handler
  */
@@ -224,6 +70,9 @@ void _XklFocusInEvHandler( XFocusChangeEvent * fev )
   Window appWin;
   XklState selectedWindowState;
 
+  if( !( _xklListenerType & XKLL_MANAGE_WINDOW_STATES ) )
+    return;
+  
   win = fev->window;
 
   switch ( fev->mode )
@@ -253,23 +102,36 @@ void _XklFocusInEvHandler( XFocusChangeEvent * fev )
   {
     if( _xklCurClient != appWin )
     {
-      Bool transparent;
+      Bool oldWinTransparent, newWinTransparent;
       XklState tmpState;
 
-      if ( XklGetState ( _xklCurClient, &tmpState ) )
-        _xklCurState = tmpState;
+      oldWinTransparent = _XklIsTransparentAppWindow( _xklCurClient );
+      if( oldWinTransparent )
+        XklDebug( 150, "Leaving transparent window\n" );
+
+      /**
+       * Reload the current state from the current window. 
+       * Do not do it for transparent window - we keep the state from 
+       * the _previous_ window.
+       */
+      if ( !oldWinTransparent && XklGetState ( _xklCurClient, &tmpState ) )
+      {
+         _XklUpdateCurState( tmpState.group, 
+                             tmpState.indicators,
+                             "Loading current (previous) state from the current (previous) window" );
+      }
 
       _xklCurClient = appWin;
       XklDebug( 150, "CurClient:changed to " WINID_FORMAT ", '%s'\n",
                 _xklCurClient, _XklGetDebugWindowTitle( _xklCurClient ) );
 
-      transparent = _XklIsTransparentAppWindow( appWin );
-      if( transparent )
+      newWinTransparent = _XklIsTransparentAppWindow( appWin );
+      if( newWinTransparent )
         XklDebug( 150, "Entering transparent window\n" );
 
-      if( XklIsGroupPerApp() == !transparent )
+      if( XklIsGroupPerApp() == !newWinTransparent )
       {
-        // We skip restoration only if we return to the same app window
+        /* We skip restoration only if we return to the same app window */
         Bool doSkip = False;
         if( _xklSkipOneRestore )
         {
@@ -302,35 +164,21 @@ void _XklFocusInEvHandler( XFocusChangeEvent * fev )
           } else
           {
             XklDebug( 150,
-                      "Both old and new focused window have state %d so no point restoring it\n",
+                      "Both old and new focused window have group %d so no point restoring it\n",
                       selectedWindowState.group );
-            _xklAllowSecondaryGroupOnce = False;
+            _XklOneSwitchToSecondaryGroupPerformed();
           }
         }
 
-#ifdef XKB_HEADERS_PRESENT
-        if( XklGetIndicatorsHandling(  ) )
+        if( ( xklVTable->features & XKLF_CAN_TOGGLE_INDICATORS ) && 
+             XklGetIndicatorsHandling(  ) )
         {
-          int i;
-          unsigned bit;
-
           XklDebug( 150,
                     "Restoring the indicators from %X to %X after gaining focus\n",
                     _xklCurState.indicators, selectedWindowState.indicators );
-          ForPhysIndicators( i,
-                             bit ) if( _xklXkb->names->indicators[i] != None )
-          {
-            Bool status;
-            status =
-              _XklSetIndicator( i,
-                                ( selectedWindowState.indicators & bit ) !=
-                                0 );
-            XklDebug( 150, "Set indicator \"%s\"/%d to %d: %d\n",
-                      _xklIndicatorNames[i], _xklXkb->names->indicators[i],
-                      selectedWindowState.indicators & bit, status );
-          }
+          _XklEnsureVTableInited();
+          (*xklVTable->xklSetIndicatorsHandler)( &selectedWindowState );
         } else
-#endif
           XklDebug( 150,
                     "Not restoring the indicators %X after gaining focus: indicator handling is not enabled\n",
                     _xklCurState.indicators );
@@ -361,6 +209,9 @@ void _XklFocusInEvHandler( XFocusChangeEvent * fev )
  */
 void _XklFocusOutEvHandler( XFocusChangeEvent * fev )
 {
+  if( !( _xklListenerType & XKLL_MANAGE_WINDOW_STATES ) )
+    return;
+  
   if( fev->mode != NotifyNormal )
   {
     XklDebug( 200,
@@ -389,12 +240,14 @@ void _XklFocusOutEvHandler( XFocusChangeEvent * fev )
     if( _XklGetAppWindow( fev->window, &p ) )
       _xklPrevAppWindow = p;
   }
-
 }
 
 /**
  * PropertyChange handler
- * Interested in WM_STATE property only
+ * Interested in :
+ *  + for XKLL_MANAGE_WINDOW_STATES
+ *    - WM_STATE property for all windows
+ *    - Configuration property of the root window
  */
 void _XklPropertyEvHandler( XPropertyEvent * pev )
 {
@@ -413,52 +266,55 @@ void _XklPropertyEvHandler( XPropertyEvent * pev )
     }
   }
 
-  if( pev->atom == _xklAtoms[WM_STATE] )
+  if( _xklListenerType & XKLL_MANAGE_WINDOW_STATES )
   {
-    Bool hasXklState = XklGetState( pev->window, NULL );
-
-    if( pev->state == PropertyNewValue )
+    if( pev->atom == _xklAtoms[WM_STATE] )
     {
-      XklDebug( 160, "New value of WM_STATE on window " WINID_FORMAT "\n",
-                pev->window );
-      if( !hasXklState )        /* Is this event the first or not? */
+      Bool hasXklState = XklGetState( pev->window, NULL );
+  
+      if( pev->state == PropertyNewValue )
       {
-        _XklAddAppWindow( pev->window, ( Window ) NULL, False,
-                          &_xklCurState );
+        XklDebug( 160, "New value of WM_STATE on window " WINID_FORMAT "\n",
+                  pev->window );
+        if( !hasXklState )        /* Is this event the first or not? */
+        {
+          _XklAddAppWindow( pev->window, ( Window ) NULL, False,
+                            &_xklCurState );
+        }
+      } else
+      {                           /* ev->xproperty.state == PropertyDelete, either client or WM can remove it, ICCCM 4.1.3.1 */
+        XklDebug( 160, "Something (%d) happened to WM_STATE of window 0x%x\n",
+                  pev->state, pev->window );
+        _XklSelectInputMerging( pev->window, PropertyChangeMask );
+        if( hasXklState )
+        {
+          XklDelState( pev->window );
+        }
       }
     } else
-    {                           /* ev->xproperty.state == PropertyDelete, either client or WM can remove it, ICCCM 4.1.3.1 */
-      XklDebug( 160, "Something (%d) happened to WM_STATE of window 0x%x\n",
-                pev->state, pev->window );
-      _XklSelectInputMerging( pev->window, PropertyChangeMask );
-      if( hasXklState )
+      if( pev->atom ==  xklVTable->baseConfigAtom
+          && pev->window == _xklRootWindow )
+    {
+      if( pev->state == PropertyNewValue )
       {
-        XklDelState( pev->window );
+        XklDebug( 160, "New value of *_NAMES_PROP_ATOM on root window\n" );
+        /* If root window got new *_NAMES_PROP_ATOM -
+         it most probably means new keyboard config is loaded by somebody */
+        _XklFreeAllInfo(  );
+        _XklLoadAllInfo(  );
       }
     }
-#ifdef XKB_HEADERS_PRESENT
-  } else
-    if( pev->atom == _xklAtoms[XKB_RF_NAMES_PROP_ATOM]
-        && pev->window == _xklRootWindow )
-  {
-    if( pev->state == PropertyNewValue )
-    {
-      XklDebug( 160, "New value of XKB_RF_NAMES_PROP_ATOM on root window\n" );
-      // If root window got new _XKB_RF_NAMES_PROP_ATOM -
-      // it most probably means new xkb is loaded by somebody
-      _XklFreeAllInfo(  );
-      _XklLoadAllInfo(  );
-    }
-#endif
-  }
+  } /* XKLL_MANAGE_WINDOW_STATES */
 }
 
 /**
- *  * CreateNotify handler. Just interested in properties and focus events...
- *   */
+ * CreateNotify handler. Just interested in properties and focus events...
+ */
 void _XklCreateEvHandler( XCreateWindowEvent * cev )
 {
-  long newmask;
+  if( !( _xklListenerType & XKLL_MANAGE_WINDOW_STATES ) )
+    return;
+
   XklDebug( 200,
             "Under-root window " WINID_FORMAT
             "/%s (%d,%d,%d x %d) is created\n", cev->window,
@@ -498,7 +354,7 @@ void _XklErrHandler( Display * dpy, XErrorEvent * evt )
     case BadWindow:
     case BadAccess:
     {
-      // in most cases this means we are late:)
+      /* in most cases this means we are late:) */
       XklDebug( 200, "ERROR: %p, " WINID_FORMAT ", %d, %d, %d\n",
                 dpy,
                 ( unsigned long ) evt->resourceid,
@@ -510,3 +366,86 @@ void _XklErrHandler( Display * dpy, XErrorEvent * evt )
       ( *_xklDefaultErrHandler ) ( dpy, evt );
   }
 }
+
+/**
+ * Some common functionality for Xkb handler
+ */
+void _XklStateModificationHandler( XklStateChange changeType, 
+                                   int grp, 
+                                   unsigned inds,
+                                   Bool setInds )
+{
+  Window focused, focusedApp;
+  XklState oldState;
+  int revert;
+  Bool haveOldState = True;
+  Bool setGroup = changeType == GROUP_CHANGED;
+
+  XGetInputFocus( _xklDpy, &focused, &revert );
+
+  if( ( focused == None ) || ( focused == PointerRoot ) )
+  {
+    XklDebug( 160, "Something with focus: " WINID_FORMAT "\n", focused );
+    return;
+  }
+
+  /** 
+   * Only if we manage states - otherwise _xklCurClient does not make sense 
+   */
+  if( !_XklGetAppWindow( focused, &focusedApp ) &&
+      _xklListenerType & XKLL_MANAGE_WINDOW_STATES )
+    focusedApp = _xklCurClient; /* what else can I do */
+
+  XklDebug( 150, "Focused window: " WINID_FORMAT ", '%s'\n", focusedApp,
+            _XklGetDebugWindowTitle( focusedApp ) );
+  if( _xklListenerType & XKLL_MANAGE_WINDOW_STATES )
+  {
+    XklDebug( 150, "CurClient: " WINID_FORMAT ", '%s'\n", _xklCurClient,
+              _XklGetDebugWindowTitle( _xklCurClient ) );
+
+    if( focusedApp != _xklCurClient )
+    {
+      /**
+       * If not state - we got the new window
+       */
+      if ( !_XklGetAppState( focusedApp, &oldState ) )
+      {
+        _XklUpdateCurState( grp, inds, 
+                            "Updating the state from new focused window" );
+        if( _xklListenerType & XKLL_MANAGE_WINDOW_STATES )
+          _XklAddAppWindow( focusedApp, ( Window ) NULL, False, &_xklCurState );
+      }
+      /**
+       * There is state - just get the state from the window
+       */
+      else
+      {
+        grp = oldState.group;
+        inds = oldState.indicators;
+      }
+      _xklCurClient = focusedApp;
+      XklDebug( 160, "CurClient:changed to " WINID_FORMAT ", '%s'\n",
+                _xklCurClient, _XklGetDebugWindowTitle( _xklCurClient ) );
+    }
+    /* If the window already has this this state - we are just restoring it!
+      (see the second parameter of stateCallback */
+    haveOldState = _XklGetAppState( _xklCurClient, &oldState );
+  } else /* just tracking the stuff, no smart things */
+  {
+    XklDebug( 160, "Just updating the current state in the tracking mode\n" );
+    memcpy( &oldState, &_xklCurState, sizeof( XklState ) );
+  }
+
+  if( setGroup || haveOldState )
+  {
+    _XklUpdateCurState( setGroup ? grp : oldState.group,
+                        setInds ? inds : oldState.indicators,
+                        "Restoring the state from the window" );
+  }
+
+  if( haveOldState )
+    _XklTryCallStateCallback( changeType, &oldState );
+
+  if( _xklListenerType & XKLL_MANAGE_WINDOW_STATES )
+    _XklSaveAppState( _xklCurClient, &_xklCurState );
+}
