@@ -1,4 +1,5 @@
 #include <time.h>
+#include <stdlib.h>
 
 #include <X11/Xatom.h>
 #include <X11/Xlib.h>
@@ -19,97 +20,27 @@ int _xklXkbEventType, _xklXkbError;
 
 static char *groupNames[XkbNumKbdGroups];
 
-const char **XklGetGroupNames( void )
+const char **_XklXkbGetGroupNames( void )
 {
   return ( const char ** ) groupNames;
 }
-#else
-const char **XklGetGroupNames( void )
+
+int _XklXkbPauseListen(  )
 {
-  static const char* dummyName = "???";
-  return ( const char ** ) &dummyName;
-}
-#endif
-
-int XklInit( Display * a_dpy )
-{
-  int opcode;
-  int scr;
-
-  if( !a_dpy )
-  {
-    XklDebug( 10, "XklInit : display is NULL ?\n");
-    return -1;
-  }
-
-  _xklDefaultErrHandler =
-    XSetErrorHandler( ( XErrorHandler ) _XklErrHandler );
-
-  _xklDpy = a_dpy;
-#ifdef XKB_HEADERS_PRESENT
-  /* Lets begin */
-  _xklXkbExtPresent = XkbQueryExtension( _xklDpy,
-                                         &opcode, &_xklXkbEventType,
-                                         &_xklXkbError, NULL, NULL );
-  if( !_xklXkbExtPresent )
-  {
-    _xklDpy = NULL;
-    XSetErrorHandler( ( XErrorHandler ) _xklDefaultErrHandler );
-    return -1;
-  }
-#endif
-
-  scr = DefaultScreen( _xklDpy );
-  _xklRootWindow = RootWindow( _xklDpy, scr );
-#ifdef XKB_HEADERS_PRESENT
-  XklDebug( 160,
-            "xkbEvenType: %X, xkbError: %X, display: %p, root: " WINID_FORMAT
-            "\n", _xklXkbEventType, _xklXkbError, _xklDpy, _xklRootWindow );
-#else
-  XklDebug( 160,
-            "NO XKB LIBS, display: %p, root: " WINID_FORMAT
-            "\n", _xklDpy, _xklRootWindow );
-#endif
-
-  _xklAtoms[WM_NAME] = XInternAtom( _xklDpy, "WM_NAME", False );
-  _xklAtoms[WM_STATE] = XInternAtom( _xklDpy, "WM_STATE", False );
-  _xklAtoms[XKLAVIER_STATE] = XInternAtom( _xklDpy, "XKLAVIER_STATE", False );
-  _xklAtoms[XKLAVIER_TRANSPARENT] =
-    XInternAtom( _xklDpy, "XKLAVIER_TRANSPARENT", False );
-#ifdef XKB_HEADERS_PRESENT
-  _xklAtoms[XKB_RF_NAMES_PROP_ATOM] =
-    XInternAtom( _xklDpy, _XKB_RF_NAMES_PROP_ATOM, False );
-  _xklAtoms[XKB_RF_NAMES_PROP_ATOM_BACKUP] =
-    XInternAtom( _xklDpy, "_XKB_RULES_NAMES_BACKUP", False );
-#endif
-
-  _xklAllowSecondaryGroupOnce = False;
-  _xklSkipOneRestore = False;
-  _xklDefaultGroup = -1;
-  _xklSecondaryGroupsMask = 0L;
-  _xklPrevAppWindow = 0;
-
-  return _XklLoadAllInfo(  )? 0 : _xklLastErrorCode;
-}
-
-int XklPauseListen(  )
-{
-#ifdef XKB_HEADERS_PRESENT
   XkbSelectEvents( _xklDpy, XkbUseCoreKbd, XkbAllEventsMask, 0 );
-//  XkbSelectEventDetails( _xklDpy,
-  //                       XkbUseCoreKbd,
-  //                     XkbStateNotify,
-  //                   0,
-  //                 0 );
+/*  XkbSelectEventDetails( _xklDpy,
+                         XkbUseCoreKbd,
+                       XkbStateNotify,
+                     0,
+                   0 );
 
-  //!!_XklSelectInput( _xklRootWindow, 0 );
-#endif
+  !!_XklSelectInput( _xklRootWindow, 0 );
+*/
   return 0;
 }
 
-int XklResumeListen(  )
+int _XklXkbResumeListen(  )
 {
-#ifdef XKB_HEADERS_PRESENT
   /* What events we want */
 #define XKB_EVT_MASK \
          (XkbStateNotifyMask| \
@@ -136,20 +67,18 @@ int XklResumeListen(  )
                          XkbUseCoreKbd,
                          XkbNamesNotify,
                          XKB_NAMES_EVT_DTL_MASK, XKB_NAMES_EVT_DTL_MASK );
-#endif
-  _XklSelectInputMerging( _xklRootWindow,
-                          SubstructureNotifyMask | PropertyChangeMask );
-  _XklGetRealState( &_xklCurState );
   return 0;
 }
 
-unsigned XklGetNumGroups(  )
+unsigned _XklXkbGetMaxNumGroups( void )
+{
+  return xklVTable->features & XKLF_MULTIPLE_LAYOUTS_SUPPORTED ?
+    XkbNumKbdGroups : 1;
+}
+
+unsigned _XklXkbGetNumGroups( void )
 {
-#ifdef XKB_HEADERS_PRESENT
   return _xklXkb->ctrls->num_groups;
-#else
-  return 1;
-#endif
 }
 
 #define KBD_MASK \
@@ -159,9 +88,16 @@ unsigned XklGetNumGroups(  )
 #define NAMES_MASK \
   ( XkbGroupNamesMask | XkbIndicatorNamesMask )
 
-void _XklFreeAllInfo(  )
+void _XklXkbFreeAllInfo(  )
 {
-#ifdef XKB_HEADERS_PRESENT
+  int i;
+  char **pi = _xklIndicatorNames;
+  for( i = 0; i < XkbNumIndicators; i++, pi++ )
+  {
+    /* only free non-empty ones */
+    if( *pi && **pi )
+      XFree( *pi );
+  }
   if( _xklXkb != NULL )
   {
     int i;
@@ -175,19 +111,18 @@ void _XklFreeAllInfo(  )
     XkbFreeKeyboard( _xklXkb, XkbAllComponentsMask, True );
     _xklXkb = NULL;
   }
-#endif
 }
 
 /**
  * Load some XKB parameters
  */
-Bool _XklLoadAllInfo(  )
+Bool _XklXkbLoadAllInfo(  )
 {
-#ifdef XKB_HEADERS_PRESENT
   int i;
-  unsigned bit;
   Atom *gna;
+  Atom *pia;
   char **groupName;
+  char **pi = _xklIndicatorNames;
 
   _xklXkb = XkbGetMap( _xklDpy, KBD_MASK, XkbUseCoreKbd );
   if( _xklXkb == NULL )
@@ -233,46 +168,41 @@ Bool _XklLoadAllInfo(  )
     return False;
   }
 
-  for( i = 0, bit = 1; i < XkbNumIndicators; i++, bit <<= 1 )
+  pia = _xklXkb->names->indicators;
+  for( i = XkbNumIndicators; --i>=0; pi++, pia++ )
   {
-    Atom a = _xklXkb->names->indicators[i];
+    Atom a = *pia;
     if( a != None )
-      _xklIndicatorNames[i] = XGetAtomName( _xklDpy, a );
+      *pi = XGetAtomName( _xklDpy, a );
     else
-      _xklIndicatorNames[i] = "";
+      *pi = "";
 
-    XklDebug( 200, "Indicator[%d] is %s\n", i, _xklIndicatorNames[i] );
+    XklDebug( 200, "Indicator[%d] is %s\n", i, *pi  );
   }
 
   XklDebug( 200, "Real indicators are %X\n",
             _xklXkb->indicators->phys_indicators );
 
-#endif
   if( _xklConfigCallback != NULL )
     ( *_xklConfigCallback ) ( _xklConfigCallbackData );
   return True;
 }
 
-void XklLockGroup( int group )
+void _XklXkbLockGroup( int group )
 {
   XklDebug( 100, "Posted request for change the group to %d ##\n", group );
-#ifdef XKB_HEADERS_PRESENT
   XkbLockGroup( _xklDpy, XkbUseCoreKbd, group );
-#endif
   XSync( _xklDpy, False );
 }
 
 /**
  * Updates current internal state from X state
  */
-void _XklGetRealState( XklState * curState_return )
+void _XklXkbGetRealState( XklState * curState_return )
 {
-#ifdef XKB_HEADERS_PRESENT
   XkbStateRec state;
-#endif
 
   curState_return->group = 0;
-#ifdef XKB_HEADERS_PRESENT
   if( Success == XkbGetState( _xklDpy, XkbUseCoreKbd, &state ) )
     curState_return->group = state.locked_group;
 
@@ -281,9 +211,7 @@ void _XklGetRealState( XklState * curState_return )
                             &curState_return->indicators ) )
     curState_return->indicators &= _xklXkb->indicators->phys_indicators;
   else
-#endif
     curState_return->indicators = 0;
-
 }
 
 /*
@@ -291,7 +219,6 @@ void _XklGetRealState( XklState * curState_return )
  */
 Bool _XklSetIndicator( int indicatorNum, Bool set )
 {
-#ifdef XKB_HEADERS_PRESENT
   XkbIndicatorMapPtr map;
 
   map = _xklXkb->indicators->maps + indicatorNum;
@@ -315,7 +242,7 @@ Bool _XklSetIndicator( int indicatorNum, Bool set )
   {
     case XkbIM_NoExplicit | XkbIM_NoAutomatic:
     {
-      // Can do nothing. Just ignore the indicator
+      /* Can do nothing. Just ignore the indicator */
       return True;
     }
 
@@ -362,9 +289,9 @@ Bool _XklSetIndicator( int indicatorNum, Bool set )
   }
 
   /* The 'which_groups' field tells when this indicator turns on
-   *      * for the 'groups' field:  base (0x1), latched (0x2), locked (0x4),
-   *           *                          or effective (0x8).
-   *                */
+   * for the 'groups' field:  base (0x1), latched (0x2), locked (0x4),
+   * or effective (0x8).
+   */
   if( map->groups )
   {
     int i;
@@ -385,13 +312,13 @@ Bool _XklSetIndicator( int indicatorNum, Bool set )
         }
       if( map->which_groups & ( XkbIM_UseLocked | XkbIM_UseEffective ) )
       {
-        // Important: Groups should be ignored here - because they are handled separately!
-        // XklLockGroup( group );
+        /* Important: Groups should be ignored here - because they are handled separately! */
+        /* XklLockGroup( group ); */
       } else if( map->which_groups & XkbIM_UseLatched )
         XkbLatchGroup( _xklDpy, XkbUseCoreKbd, group );
       else
       {
-        // Can do nothing. Just ignore the indicator
+        /* Can do nothing. Just ignore the indicator */
         return True;
       }
     } else
@@ -440,6 +367,103 @@ Bool _XklSetIndicator( int indicatorNum, Bool set )
       return True;
     }
   }
-#endif
   return True;
 }
+
+#endif
+
+int _XklXkbInit( void )
+{
+#ifdef XKB_HEADERS_PRESENT
+  int opcode;
+  static XklVTable xklXkbVTable =
+  {
+    "XKB",
+    XKLF_CAN_TOGGLE_INDICATORS |
+      XKLF_CAN_OUTPUT_CONFIG_AS_ASCII |
+      XKLF_CAN_OUTPUT_CONFIG_AS_BINARY,
+    _XklXkbConfigActivate,
+    _XklXkbConfigInit,
+    _XklXkbConfigLoadRegistry,
+    _XklXkbConfigWriteFile,
+    _XklXkbEventHandler,
+    _XklXkbFreeAllInfo,
+    _XklXkbGetGroupNames,
+    _XklXkbGetMaxNumGroups,
+    _XklXkbGetNumGroups,
+    _XklXkbGetRealState,
+    _XklXkbLoadAllInfo,
+    _XklXkbLockGroup,
+    _XklXkbPauseListen,
+    _XklXkbResumeListen,
+    _XklXkbSetIndicators,
+  };
+
+  if( getenv( "XKL_XKB_DISABLE" ) != NULL )
+    return -1;
+
+  _xklXkbExtPresent = XkbQueryExtension( _xklDpy,
+                                         &opcode, &_xklXkbEventType,
+                                         &_xklXkbError, NULL, NULL );
+  if( !_xklXkbExtPresent )
+  {
+    XSetErrorHandler( ( XErrorHandler ) _xklDefaultErrHandler );
+    return -1;
+  }
+  
+  XklDebug( 160,
+            "xkbEvenType: %X, xkbError: %X, display: %p, root: " WINID_FORMAT
+            "\n", _xklXkbEventType, _xklXkbError, _xklDpy, _xklRootWindow );
+
+  xklXkbVTable.baseConfigAtom =
+    XInternAtom( _xklDpy, _XKB_RF_NAMES_PROP_ATOM, False );
+  xklXkbVTable.backupConfigAtom =
+    XInternAtom( _xklDpy, "_XKB_RULES_NAMES_BACKUP", False );
+
+  xklXkbVTable.defaultModel = "pc101";
+  xklXkbVTable.defaultLayout = "us";
+
+  xklVTable = &xklXkbVTable;
+
+  /* First, we have to assign xklVTable - 
+     because this function uses it */
+  
+  if( _XklXkbConfigMultipleLayoutsSupported() )
+    xklXkbVTable.features |= XKLF_MULTIPLE_LAYOUTS_SUPPORTED;
+  
+  return 0;
+#else
+  XklDebug( 160,
+            "NO XKB LIBS, display: %p, root: " WINID_FORMAT
+            "\n", _xklDpy, _xklRootWindow );
+  return -1;
+#endif
+}
+
+#ifdef XKB_HEADERS_PRESENT
+const char *_XklXkbGetXkbEventName( int xkb_type )
+{
+  /* Not really good to use the fact of consecutivity
+     but XKB protocol extension is already standartized so... */
+  static const char *evtNames[] = {
+    "XkbNewKeyboardNotify",
+    "XkbMapNotify",
+    "XkbStateNotify",
+    "XkbControlsNotify",
+    "XkbIndicatorStateNotify",
+    "XkbIndicatorMapNotify",
+    "XkbNamesNotify",
+    "XkbCompatMapNotify",
+    "XkbBellNotify",
+    "XkbActionMessage",
+    "XkbAccessXNotify",
+    "XkbExtensionDeviceNotify",
+    "LASTEvent"
+  };
+  xkb_type -= XkbNewKeyboardNotify;
+  if( xkb_type < 0 || 
+      xkb_type >= ( sizeof( evtNames ) / sizeof( evtNames[0] ) ) )
+    return "UNKNOWN";
+  return evtNames[xkb_type];
+}
+#endif
